#!/usr/bin/env python3
"""
Router Config Backup
Automated backup tool for network device configurations.
Parameters:
    - devices: Path to device list file (CSV format)
    - output: Directory to store backups
    - protocol: Protocol to use (ssh/telnet)
Requires: netmiko, paramiko packages
"""
import netmiko
import paramiko
import argparse
import logging
import csv
import os
from datetime import datetime
import threading
import queue
import json

class RouterConfigBackup:
    def __init__(self, devices_file, output_dir, protocol='ssh'):
        self.devices_file = devices_file
        self.output_dir = output_dir
        self.protocol = protocol.lower()
        self.backup_queue = queue.Queue()
        self.results = []
        self.setup_logging()

    def setup_logging(self):
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s - %(levelname)s - %(message)s'
        )
        self.logger = logging.getLogger(__name__)

    def read_devices(self):
        devices = []
        try:
            with open(self.devices_file, 'r') as f:
                reader = csv.DictReader(f)
                for row in reader:
                    devices.append({
                        'host': row['host'],
                        'device_type': row.get('device_type', 'cisco_ios'),
                        'username': row['username'],
                        'password': row['password'],
                        'secret': row.get('secret', ''),
                        'port': int(row.get('port', 22 if self.protocol == 'ssh' else 23))
                    })
            return devices
        except Exception as e:
            self.logger.error(f"Error reading devices file: {str(e)}")
            return []

    def backup_device(self, device):
        try:
            self.logger.info(f"Connecting to {device['host']}")
            
            # Create connection handler
            connection = netmiko.ConnectHandler(
                device_type=device['device_type'],
                host=device['host'],
                username=device['username'],
                password=device['password'],
                secret=device['secret'],
                port=device['port']
            )

            # Enter enable mode if required
            if device['secret']:
                connection.enable()

            # Get device configuration
            config = connection.send_command('show running-config')
            hostname = connection.send_command('show hostname').strip()
            
            # Create backup filename
            timestamp = datetime.now().strftime('%Y%m%d_%H%M%S')
            filename = f"{hostname}_{timestamp}_backup.txt"
            filepath = os.path.join(self.output_dir, filename)

            # Save configuration
            with open(filepath, 'w') as f:
                f.write(config)

            # Get additional device information
            version = connection.send_command('show version')
            interfaces = connection.send_command('show interfaces')

            # Save metadata
            metadata = {
                'timestamp': timestamp,
                'hostname': hostname,
                'device_type': device['device_type'],
                'backup_file': filename,
                'version_info': version.split('\n')[0],
                'interface_count': len(interfaces.split('interface')),
                'config_size': len(config)
            }

            metadata_file = f"{hostname}_{timestamp}_metadata.json"
            metadata_path = os.path.join(self.output_dir, metadata_file)
            
            with open(metadata_path, 'w') as f:
                json.dump(metadata, f, indent=4)

            connection.disconnect()
            
            self.results.append({
                'host': device['host'],
                'status': 'success',
                'backup_file': filename,
                'metadata_file': metadata_file
            })
            
            self.logger.info(f"Backup completed for {device['host']}")
            
        except Exception as e:
            self.logger.error(f"Error backing up {device['host']}: {str(e)}")
            self.results.append({
                'host': device['host'],
                'status': 'failed',
                'error': str(e)
            })

    def verify_backup(self, backup_file):
        try:
            with open(backup_file, 'r') as f:
                content = f.read()
                if len(content) > 0 and 'hostname' in content.lower():
                    return True
            return False
        except Exception:
            return False

    def generate_report(self):
        report = {
            'timestamp': datetime.now().isoformat(),
            'summary': {
                'total_devices': len(self.results),
                'successful_backups': len([r for r in self.results if r['status'] == 'success']),
                'failed_backups': len([r for r in self.results if r['status'] == 'failed'])
            },
            'details': self.results
        }

        report_file = os.path.join(self.output_dir, 
                                 f"backup_report_{datetime.now().strftime('%Y%m%d_%H%M%S')}.json")
        
        try:
            with open(report_file, 'w') as f:
                json.dump(report, f, indent=4)
            self.logger.info(f"Backup report saved to {report_file}")
        except Exception as e:
            self.logger.error(f"Error saving backup report: {str(e)}")

    def start_backup(self, max_threads=5):
        # Create output directory if it doesn't exist
        os.makedirs(self.output_dir, exist_ok=True)

        devices = self.read_devices()
        if not devices:
            self.logger.error("No devices found in the devices file")
            return

        threads = []
        for device in devices:
            while threading.active_count() > max_threads:
                time.sleep(1)
            
            thread = threading.Thread(target=self.backup_device, args=(device,))
            thread.start()
            threads.append(thread)

        # Wait for all backups to complete
        for thread in threads:
            thread.join()

        self.generate_report()
        
        self.logger.info("\nBackup Summary:")
        self.logger.info(f"Total devices: {len(self.results)}")
        self.logger.info(f"Successful backups: {len([r for r in self.results if r['status'] == 'success'])}")
        self.logger.info(f"Failed backups: {len([r for r in self.results if r['status'] == 'failed'])}")

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Router Config Backup")
    parser.add_argument("devices", help="Path to device list file (CSV format)")
    parser.add_argument("--output", default="backups",
                      help="Directory to store backups")
    parser.add_argument("--protocol", choices=['ssh', 'telnet'],
                      default='ssh', help="Protocol to use")
    parser.add_argument("--threads", type=int, default=5,
                      help="Maximum number of concurrent backups")
    args = parser.parse_args()

    backup = RouterConfigBackup(args.devices, args.output, args.protocol)
    backup.start_backup(args.threads)
